#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ovsdbapp.backend.ovs_idl import command as cmd
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp.backend.ovs_idl import rowview


class TsAddCommand(cmd.AddCommand):
    table_name = 'Transit_Switch'

    def __init__(self, api, switch, may_exist=False, **columns):
        super().__init__(api)
        self.switch = switch
        self.columns = columns
        self.may_exist = may_exist

    def run_idl(self, txn):
        # There is requirement for name to be unique
        # (index in ovn-ic-nb.ovsschema)
        switch = idlutils.row_by_value(self.api.idl, self.table_name, 'name',
                                       self.switch, None)
        if switch:
            if self.may_exist:
                self.result = rowview.RowView(switch)
                return
            raise RuntimeError("Transit Switch %s exists" % self.switch)
        switch = txn.insert(self.api.tables[self.table_name])
        switch.name = self.switch
        self.set_columns(switch, **self.columns)
        self.result = switch.uuid


class TsDelCommand(cmd.BaseCommand):
    def __init__(self, api, switch, if_exists=False):
        super().__init__(api)
        self.switch = switch
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            switch = self.api.lookup('Transit_Switch', self.switch)
            switch.delete()
        except idlutils.RowNotFound as e:
            if self.if_exists:
                return
            msg = "Transit Switch %s does not exist" % self.switch
            raise RuntimeError(msg) from e


class TsListCommand(cmd.ReadOnlyCommand):
    def run_idl(self, txn):
        table = self.api.tables['Transit_Switch']
        self.result = [rowview.RowView(r) for r in table.rows.values()]


class TsGetCommand(cmd.BaseGetRowCommand):
    table = 'Transit_Switch'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ovsdbapp.backend.ovs_idl import command as cmd
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp.backend.ovs_idl import rowview


class TableGlobalIsEmpty(idlutils.RowNotFound):
    message = "Table 'Global' is empty"


def get_global_record(api):
    # there should be only one record in 'Global' table
    try:
        return next((r for r in api.tables['Global'].rows.values()))
    except StopIteration as e:
        raise TableGlobalIsEmpty from e


class _ListCommand(cmd.ReadOnlyCommand):
    def run_idl(self, txn):
        table = self.api.tables[self.table_name]
        self.result = [rowview.RowView(r) for r in table.rows.values()]


class AddPsCommand(cmd.AddCommand):
    table_name = 'Physical_Switch'

    def __init__(self, api, pswitch, may_exist=False, **columns):
        super().__init__(api)
        self.pswitch = pswitch
        self.columns = columns
        self.may_exist = may_exist

    def run_idl(self, txn):
        config = get_global_record(self.api)
        pswitch = idlutils.row_by_value(self.api.idl, self.table_name, 'name',
                                        self.pswitch, None)
        if pswitch:
            if self.may_exist:
                self.result = rowview.RowView(pswitch)
                return
            msg = "Physical switch %s exists" % self.pswitch
            raise RuntimeError(msg)
        pswitch = txn.insert(self.api.tables[self.table_name])
        pswitch.name = self.pswitch
        self.set_columns(pswitch, **self.columns)
        config.addvalue('switches', pswitch)
        self.result = pswitch.uuid


class DelPsCommand(cmd.BaseCommand):
    table_name = 'Physical_Switch'

    def __init__(self, api, pswitch, if_exists=False):
        super().__init__(api)
        self.pswitch = pswitch
        self.if_exists = if_exists

    def run_idl(self, txn):
        config = get_global_record(self.api)
        try:
            pswitch = self.api.lookup(self.table_name, self.pswitch)
            config.delvalue('switches', pswitch)
            pswitch.delete()
        except idlutils.RowNotFound as e:
            if self.if_exists:
                return
            msg = "Physical switch %s does not exist" % self.pswitch
            raise RuntimeError(msg) from e


class ListPsCommand(_ListCommand):
    table_name = 'Physical_Switch'


class GetPsCommand(cmd.BaseGetRowCommand):
    table = 'Physical_Switch'


class AddPortCommand(cmd.AddCommand):
    table_name = 'Physical_Port'

    def __init__(self, api, pswitch, port, may_exist=False):
        super().__init__(api)
        self.pswitch = pswitch
        self.port = port
        self.may_exist = may_exist
        self.conditions = [('name', '=', self.port)]

    def run_idl(self, txn):
        pswitch = self.api.lookup('Physical_Switch', self.pswitch)
        port = next((p for p in pswitch.ports
                     if idlutils.row_match(p, self.conditions)), None)
        if port:
            if self.may_exist:
                self.result = rowview.RowView(port)
                return
            msg = "Physical port %s exists in %s" % (self.port, self.pswitch)
            raise RuntimeError(msg)
        port = txn.insert(self.api.tables[self.table_name])
        port.name = self.port
        pswitch.addvalue('ports', port)
        self.result = port.uuid


class DelPortCommand(cmd.BaseCommand):
    table_name = 'Physical_Port'

    def __init__(self, api, pswitch, port, if_exists=False):
        super().__init__(api)
        self.pswitch = pswitch
        self.port = port
        self.if_exists = if_exists
        self.conditions = [('name', '=', self.port)]

    def run_idl(self, txn):
        pswitch = self.api.lookup('Physical_Switch', self.pswitch)
        port = next((p for p in pswitch.ports
                     if idlutils.row_match(p, self.conditions)), None)
        if not port:
            if self.if_exists:
                return
            msg = "Physical port %s does not exist in %s" % (self.port,
                                                             self.pswitch)
            raise RuntimeError(msg)
        pswitch.delvalue('ports', port)
        port.delete()


class ListPortsCommand(cmd.ReadOnlyCommand):
    table_name = 'Physical_Switch'

    def __init__(self, api, pswitch):
        super().__init__(api)
        self.pswitch = pswitch

    def run_idl(self, txn):
        pswitch = self.api.lookup(self.table_name, self.pswitch)
        self.result = [rowview.RowView(port) for port in pswitch.ports]


class GetPortCommand(cmd.BaseGetRowCommand):
    table = 'Physical_Port'


class AddLsCommand(cmd.AddCommand):
    table_name = 'Logical_Switch'

    def __init__(self, api, switch, may_exist=False, **columns):
        super().__init__(api)
        self.switch = switch
        self.columns = columns
        self.may_exist = may_exist

    def run_idl(self, txn):
        switch = idlutils.row_by_value(self.api.idl, self.table_name, 'name',
                                       self.switch, None)
        if switch:
            if self.may_exist:
                self.result = rowview.RowView(switch)
                return
            msg = "Logical switch %s exists" % self.switch
            raise RuntimeError(msg)
        switch = txn.insert(self.api.tables[self.table_name])
        switch.name = self.switch
        self.set_columns(switch, **self.columns)
        self.result = switch.uuid


class DelLsCommand(cmd.BaseCommand):
    table_name = 'Logical_Switch'

    def __init__(self, api, switch, if_exists=False):
        super().__init__(api)
        self.switch = switch
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            entity = self.api.lookup(self.table_name, self.switch)
            entity.delete()
        except idlutils.RowNotFound as e:
            if self.if_exists:
                return
            msg = "Logical switch %s does not exist" % self.switch
            raise RuntimeError(msg) from e


class ListLsCommand(_ListCommand):
    table_name = 'Logical_Switch'


class GetLsCommand(cmd.BaseGetRowCommand):
    table = 'Logical_Switch'


class BindLsCommand(cmd.BaseCommand):
    table_name = 'Physical_Port'

    def __init__(self, api, pswitch, port, vlan, switch):
        super().__init__(api)
        self.pswitch = pswitch
        self.port = port
        self.vlan = vlan
        self.switch = switch
        self.conditions = [('name', '=', self.port)]

    def run_idl(self, txn):
        pswitch = self.api.lookup('Physical_Switch', self.pswitch)
        switch = self.api.lookup('Logical_Switch', self.switch)
        port = next((p for p in pswitch.ports
                     if idlutils.row_match(p, self.conditions)), None)
        if not port:
            raise idlutils.RowNotFound(table=self.table_name,
                                       col='name', match=self.port)
        port.setkey('vlan_bindings', self.vlan, switch.uuid)


class UnbindLsCommand(cmd.BaseCommand):
    table_name = 'Physical_Port'

    def __init__(self, api, pswitch, port, vlan):
        super().__init__(api)
        self.pswitch = pswitch
        self.port = port
        self.vlan = vlan
        self.conditions = [('name', '=', self.port)]

    def run_idl(self, txn):
        pswitch = self.api.lookup('Physical_Switch', self.pswitch)
        port = next((p for p in pswitch.ports
                     if idlutils.row_match(p, self.conditions)), None)
        if not port:
            raise idlutils.RowNotFound(table=self.table_name,
                                       col='name', match=self.port)
        port.delkey('vlan_bindings', self.vlan)


class _ClearMacsCommand(cmd.BaseCommand):
    def __init__(self, api, switch):
        super().__init__(api)
        self.switch = switch

    def run_idl(self, txn):
        switch = self.api.lookup('Logical_Switch', self.switch)
        macs = []

        for table_name in self.table_names:
            macs.extend(idlutils.rows_by_value(self.api.idl,
                                               table_name,
                                               'logical_switch', switch))
        for mac in macs:
            mac.delete()


class ClearLocalMacsCommand(_ClearMacsCommand):
    table_names = ['Ucast_Macs_Local', 'Mcast_Macs_Local']


class ClearRemoteMacsCommand(_ClearMacsCommand):
    table_names = ['Ucast_Macs_Remote', 'Mcast_Macs_Remote']


class _ListMacsCommand(cmd.ReadOnlyCommand):
    def __init__(self, api, switch):
        super().__init__(api)
        self.switch = switch

    def run_idl(self, txn):
        switch = self.api.lookup('Logical_Switch', self.switch)
        self.result = []

        for table_name in self.table_names:
            self.result.append([rowview.RowView(mac)
                                for mac in idlutils.rows_by_value(
                                    self.api.idl, table_name,
                                    'logical_switch', switch)])


class ListLocalMacsCommand(_ListMacsCommand):
    table_names = ['Ucast_Macs_Local', 'Mcast_Macs_Local']


class ListRemoteMacsCommand(_ListMacsCommand):
    table_names = ['Ucast_Macs_Remote', 'Mcast_Macs_Remote']



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ovsdbapp.backend import ovs_idl
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp.schema.ovn_ic_northbound import api
from ovsdbapp.schema.ovn_ic_northbound import commands as cmd


class OvnIcNbApiIdlImpl(ovs_idl.Backend, api.API):
    schema = 'OVN_IC_Northbound'
    lookup_table = {
        'Transit_Switch': idlutils.RowLookup('Transit_Switch', 'name', None),
    }

    def ts_add(self, switch, may_exist=False, **columns):
        return cmd.TsAddCommand(self, switch, may_exist, **columns)

    def ts_del(self, switch, if_exists=False):
        return cmd.TsDelCommand(self, switch, if_exists)

    def ts_list(self):
        return cmd.TsListCommand(self)

    def ts_get(self, switch):
        return cmd.TsGetCommand(self, switch)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import atexit
import itertools
import logging
import queue
import threading

# Several supported distros don't package sortedcontainers. For this reason
# python-ovs, which requires it for indexing support, has an in-tree copy of
# it. We already require sortedcontainers through python-ovs' dependency, so
# to avoid issues with distro packaging, we don't add sortedcontainers as a
# hard direct requirement of ovsdbapp, and use python-ovs' if it is missing.
try:
    import sortedcontainers
except ImportError:
    from ovs.compat import sortedcontainers

LOG = logging.getLogger(__name__)
STOP_EVENT = ("STOP", None, None, None)


class RowEvent(object, metaclass=abc.ABCMeta):
    ROW_CREATE = "create"
    ROW_UPDATE = "update"
    ROW_DELETE = "delete"
    ONETIME = False
    event_name = 'RowEvent'
    priority = 20

    def __init__(self, events, table, conditions, old_conditions=None):
        self.table = table
        self.events = events
        self.conditions = conditions
        self.old_conditions = old_conditions

    @property
    def key(self):
        return (self.__class__, self.table, tuple(self.events))

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        try:
            return (self.key == other.key and
                    self.conditions == other.conditions and
                    self.priority == other.priority)
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return ("%s(events=%r, table=%r, conditions=%r, old_conditions=%r), "
                "priority=%d" %
                (self.__class__.__name__, self.events,
                 self.table, self.conditions, self.old_conditions,
                 self.priority))

    @abc.abstractmethod
    def matches(self, event, row, old=None):
        """Test that `event` on `row` matches watched events

        :param event: event type
        :type event:  ROW_CREATE, ROW_UPDATE, or ROW_DELETE
        :param row:
        :param old:
        :returns:    boolean, True if match else False
        """

    @abc.abstractmethod
    def run(self, event, row, old):
        """Method to run when the event matches"""


class WaitEvent(RowEvent):
    event_name = 'WaitEvent'
    ONETIME = True
    priority = 10

    def __init__(self, *args, **kwargs):
        self.event = threading.Event()
        self.timeout = kwargs.pop('timeout', None)
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def matches(self, event, row, old=None):
        """Test that `event on `row` matches watched events. See: RowEvent"""

    def run(self, event, row, old):
        self.event.set()

    def wait(self):
        return self.event.wait(self.timeout)


class RowEventHandler(object):
    def __init__(self):
        self._queues = sortedcontainers.SortedDict(lambda p: -p)
        self._lock = threading.Lock()
        self.notifications = queue.Queue()
        self.notify_thread = threading.Thread(target=self.notify_loop)
        self.notify_thread.daemon = True
        atexit.register(self.shutdown)
        self.start()

    def start(self):
        self.notify_thread.start()

    def _get_queue(self, event):
        return self._queues.setdefault(event.priority, set())

    @property
    def _watched_events(self):
        return itertools.chain(*self._queues.values())

    def _add(self, event):
        self._get_queue(event).add(event)

    def _discard(self, event):
        self._get_queue(event).discard(event)

    @staticmethod
    def match(candidate, event, row, updates):
        try:
            return candidate.matches(event, row, updates)
        except Exception:
            LOG.exception("Event not matched due to this exception:\n")
            return False

    def matching_events(self, event, row, updates):
        with self._lock:
            return tuple(t for t in self._watched_events
                         if self.match(t, event, row, updates))

    def watch_event(self, event):
        with self._lock:
            self._add(event)

    def watch_events(self, events):
        with self._lock:
            for event in events:
                self._add(event)

    def unwatch_event(self, event):
        with self._lock:
            self._discard(event)

    def unwatch_events(self, events):
        with self._lock:
            for event in events:
                self._discard(event)

    def shutdown(self):
        self.notifications.put(STOP_EVENT)

    def notify_loop(self):
        while True:
            try:
                match, event, row, updates = self.notifications.get()
                if (match, event, row, updates) == STOP_EVENT:
                    self.notifications.task_done()
                    break
                match.run(event, row, updates)
                if match.ONETIME:
                    self.unwatch_event(match)
                self.notifications.task_done()
            except Exception:
                # If any unexpected exception happens we don't want the
                # notify_loop to exit.
                LOG.exception('Unexpected exception in notify_loop')

    def notify(self, event, row, updates=None):
        """Method for calling backend to call for each DB update

        :param event:   Backend representation of event type, e.g.
                        create, update, delete
        :param row:     Backend representation of a Row object. If it is not
                        immutable, it should be converted or guaranteed not to
                        be changed in other threads.
        :param updates: Backend representation of updates to a Row. e.g.
                        a Row object with just changed attributes, a
                        dictionary of changes, etc.
        """
        matching = self.matching_events(
            event, row, updates)
        for match in matching:
            self.notifications.put((match, event, row, updates))
